#-----------------------------------------------------------------------------
# BVH Mixer
# 9/6/04
# Revised slightly 3/4/07 to work in P7

# Copyright (c) 2004 David G. Drumright (ockham)
# Pick a BVH file.  Assign it a startframe and a multiplier.
# The listbox carries the parts that the BVH wants to affect;
# you can pick some or all.
# Hit go and we try to mix it into current figure.
#-----------------------------------------------------------------------------

from Tkinter import *
root = Tk()
import tkFileDialog
import poser
import string
import os

#-----------------------------------------------------------------------------

global FileFrameCount
global FileFrameLength
global Guide
global Values
global TheFig

scene=poser.Scene()

FileFrameCount=0
FileFrameLength=0.0
TheFig = scene.CurrentFigure()

Guide=[]  # Guide array includes one entry for every channel in the file.

Values=[] # Repeats a parallel of the Guide for each frame.
# So, when reading the lower part of the file, we simply fill up the Values array.
# Then, when we apply the Values, we use the Guide; for each frame,
# go through Guide to get the actor and parm, while we run straight through Values.

# Translate the BVH names into Poser parm codes, with an extra flag to distinguish Tran from Rot quickly
ParmDict={  "Xposition":(poser.kParmCodeXTRAN,0),
            "Yposition":(poser.kParmCodeYTRAN,0),
            "Zposition":(poser.kParmCodeZTRAN,0),
            "Xrotation":(poser.kParmCodeXROT,1),
            "Yrotation":(poser.kParmCodeYROT,1),
            "Zrotation":(poser.kParmCodeZROT,1),
            "Xscale":   (poser.kParmCodeXSCALE,2),
            "Yscale":   (poser.kParmCodeYSCALE,2),
            "Zscale":   (poser.kParmCodeZSCALE,2) }
# Dict to translate the BVH names to poser parm codes
scene = poser.Scene()
BasePathUncut=poser.AppLocation()
BasePath=BasePathUncut[:(-len('poser.exe'))]
MyPath=os.path.normpath(BasePath+'/runtime/*.BVH') # Use this for release
PZ2Path=os.path.normpath(BasePath+'/runtime/libraries/pose')

MainScaler = (1.0/90.0)
# 94.974 is the approximate scaler that Poser uses for DCF.  Thigh length is .203177
# 86.843  for Misaki.  Thigh length is .172864
# I think 500 * Thigh length might be a good scaler?
# For now, just use 90, then fill in later if more info.

#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------

'''

Reading the file.

1. Read upper section, which associates figure's parts and parameters with an ordinal.

The file starts like this:

HIERARCHY
ROOT hip
{
    OFFSET  0.00  0.00  0.00
    CHANNELS 6 Xposition Yposition Zposition Xrotation Zrotation Yrotation
    JOINT abdomen
    {
        OFFSET  0.146277 2.494683 -0.237917
        CHANNELS 3 Xrotation Zrotation Yrotation
        JOINT chest
        {
We can ignore the brackets. Basically we need to look first for
ROOT %s for the central part (which isn't really any different from the others
in this context) then CHANNELS %d [ same number of parm names as the number here ]

After the ROOT, we look for JOINT %s followed by CHANNELS %d [ parms ]
During this loop, we build a list with (name,parm) tuples as elements.  Since
the ordinal just increments through here, we don't need a dict; the list
will naturally be indexed properly.

The start of the second part is signaled by MOTION so we can use that
as the break-check for first loop.
MOTION
Frames:     2
Frame Time: 0.033333
-0.000031 40.332233 0.278907 0.000000 ...............

This lower part has no linebreaks, so we'll have to read it
one number at a time.  Otherwise the fileline could get
monstrous....

'''
#-----------------------------------------------------------------------------
def AltReadLine(fp): # Helper to read any sort of file format
    BuiltLine=''
    while 1:
        newchar=fp.read(1)
        if not newchar: return None # EOF
        elif newchar=='\r' or newchar=='\n':
            # We have a newline.  Check the next char.
            # If this one is \r and next is \n, we have
            # a CR LF sequence.  Return the line and let
            # the file pointer run naturally.
            # Otherwise, the next char is the beginning
            # of the next line, so we need to back up one
            # before returning.
            AheadChar=fp.read(1)
            if newchar =='\r' and AheadChar =='\n':
                return BuiltLine+'\n'
            else:
                fp.seek(-1,1) # back up
                return BuiltLine+'\n'
        else:
            BuiltLine=BuiltLine+newchar
#-----------------------------------------------------------------------------
def FakeFrames(DesiredCount):
    # Write a short PZ2 file forcing the desired number of
    # frames; load it; delete it.
    PZ2Name=os.path.normpath(PZ2Path+'/zzztemp.pz2')
    fpFake=open(PZ2Name,"wt")
    if not fpFake: return 0
    fpFake.write('{\nversion\n\t{\n\tnumber 4.2\n\t}\nmovieInfo\n\t{\n\t')
    fpFake.write('numFrames %d\n\t}\n}' % FileFrameCount)
    fpFake.close()
    try:
        scene.LoadLibraryPose(PZ2Name)
    except:
        return 0
    os.remove(PZ2Name)
    return DesiredCount
#-----------------------------------------------------------------------------

def ReadOneNumber(fp): # Helper to read one number.
    # assume fp is already open
    NumSt = ''
    ch = ''
    # We read one char at a time until we hit a delimiter.
    # We also need to account for double delimiters. Sometimes
    # a newline is inserted between the sequences for each frame,
    # so we may have \t\n.
    WeHaveNum=0
    while 1:
        ch = fp.read(1)
        if ch=='': return None # break  # Empty string means EOF, but take last number
        elif not ch: return None  # None means EOF?
        # Take all possible delimiters:
        elif (ch in string.whitespace):
            if WeHaveNum: break # After numbers: stop and figure it.
            else: continue # Implied else: Still jumping across extra delims
        else:
            NumSt=NumSt+ch
            WeHaveNum=1
    return float(string.lstrip(NumSt))

#-----------------------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):

        self.master = master
        master.title("BVH mixer")
        self.FloorVar=IntVar()
        self.FloorVar.set(1)
        self.RelVar=IntVar()
        self.RelVar.set(1)


        self.StatusEntry = Entry(self.master,width=60)
        self.StatusEntry.grid(row=0,column=0,columnspan=2)
        self.StatusEntry.insert(0,"Figure %s is selected. Pick a BVH file." % TheFig.Name())

        # Main elements
        self.ButtonPick = Button(self.master, text="Pick file", command=self.ReadOneBVH)
        self.ButtonPick.grid(row=1, column=0,columnspan=2,sticky=W+E)
        self.ListFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ListFrame.grid(  row=2,column=0,pady=1)
        self.EntryFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
        self.EntryFrame.grid(row=2,column=1,pady=1)

        # List and scroll, in ListFrame
        self.PartScroll  = Scrollbar(self.ListFrame, orient=VERTICAL)
        self.PartScroll.grid( row=0, column=0,sticky=N+S+E)
        self.PartBox  = Listbox(self.ListFrame, height=15, selectmode=MULTIPLE,yscrollcommand = self.PartScroll.set)
        self.PartBox.grid( row=0, column=1)
        self.PartScroll["command"] = self.PartBox.yview

        # Entry boxes in EntryFrame
        Label(self.EntryFrame,text="Start frame").grid(row=0,column=0)
        self.StartEntry=Entry(self.EntryFrame,width=8)
        self.StartEntry.grid(row=0,column=1,sticky=W,pady=3)
        self.StartEntry.insert(0,str(1))

        Label(self.EntryFrame,text="Rot scale").grid(row=1,column=0)
        self.RotScaleEntry=Entry(self.EntryFrame,width=8)
        self.RotScaleEntry.grid(row=1,column=1,sticky=W,pady=3)

        Label(self.EntryFrame,text="Tran scale").grid(row=2,column=0)
        self.TrnScaleEntry=Entry(self.EntryFrame,width=8)
        self.TrnScaleEntry.grid(row=2,column=1,sticky=W,pady=3)

        Label(self.EntryFrame,text="How many times:").grid(row=3,column=0)
        self.RepEntry=Entry(self.EntryFrame,width=8)
        self.RepEntry.grid(row=3,column=1,sticky=W,pady=3)
        self.RepEntry.insert(0,str(1))

        Label(self.EntryFrame,text="Drop to floor:").grid(row=4,column=0)
        self.FloorCheck=Checkbutton(self.EntryFrame,var=self.FloorVar)
        self.FloorCheck.grid(row=4,column=1)

        Label(self.EntryFrame,text="Relative:").grid(row=4,column=2)
        self.RelCheck=Checkbutton(self.EntryFrame,var=self.RelVar)
        self.RelCheck.grid(row=4,column=3)


        self.ButtonGo = Button(self.master, text="Go!", command=self.HandleGo)
        self.ButtonGo.grid(row=3, column=0)

        self.ButtonCancel = Button(self.master, text="Close", command=self.HandleCancel)
        self.ButtonCancel.grid(row=3, column=1)

        self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

        # Check for a figure first.
        if not TheFig:
            self.StatusEntry.delete(0,END)
            self.StatusEntry.insert(0,"No figure selected.  Cancel and restart.")

    # - - - - - - - - - - - - - - - - - -

    def ReadOneBVH(self):
        global FileFrameCount
        global FileFrameLength
        global Guide
        global Values

        # Pop up file box to get a BVH.
        fn=tkFileDialog.askopenfilename(initialfile=MyPath, filetypes=[("BVH files", "*.BVH *.bvh")] )

        if not fn:
            self.ShowStatus("No file selected. Try again.")
            return 0

        # Open the file
        FullFileName=os.path.abspath(fn)
        fp=open(FullFileName,'r')
        if not fp: return 0
        # Get base name for title
        BaseName=os.path.basename(fn)
        self.master.title("BVH Mixer: %s" % BaseName)

        # First loop. Look for ROOT %s then CHANNELS %d names......
        LookingForChannels=0
        while 1:
            st = AltReadLine(fp)

            if not st:    break # EOF

            Parts=string.split(st)
            LP=len(Parts)
            if LP < 1:  # only if we have a pure newline
                continue
            elif string.count(Parts[0],'ROOT'):  # ROOT and JOINT are identical for our purposes.
                # The name may be multi-word as in Right Foot or single as in rFoot, so
                # we need to join the parts after ROOT or JOINT.
                LookingForChannels=1
                Name = string.join(Parts[1:])
            elif string.count(Parts[0],'JOINT'):
                LookingForChannels=1
                Name = string.join(Parts[1:])
            elif (LP > 2) and LookingForChannels and string.count(Parts[0],'CHANNELS'):
                LookingForChannels=0 # Unset the flag
                # Read the parms, add each to the list. Note that Parts[1] is the number of chans;
                # we don't use it in this parsing but we need it for the second read-loop.
                for OneParm in Parts[2:]:
                    # Have to adjust the BVH names to ours.
                    (PoserParm,ParmType)=ParmDict[OneParm]
                    Guide.append( (Name,PoserParm,ParmType) )
            elif (LP > 1) and string.count(Parts[0],'Frames:'): # Line like  Frames: 2
                FileFrameCount = int(Parts[1])
            elif (LP > 2) and string.count(Parts[1],'Time:'):  # Line like  Frame Time: 0.033333
                FileFrameLength = float(Parts[2])
                break # This is the last of the 'text' reads.

        # After we hit Frame Time, continue reading in a different way, to get the numbers.
        # At this stage we just fill Values with the numbers; we will match this with
        # the Guide when we actually apply.
        while 1:
            Num = ReadOneNumber(fp)
            if Num == None: break
            else:
                Values.append(Num)

        fp.close()  # Close the file.

        # Take the names and fill the listbox with actor names.
        for (ActorName,ParmCode,ParmType) in Guide:
            # Run through the existing items to see if this one is already there.
            # (Would be nicer to use the dict-trick for nonredundancy, but in this
            # case we need to preserve the original order, which isn't alphabetical.)
            Found=0
            for Index in range(self.PartBox.size()):
                ts=self.PartBox.get(Index)
                if ts == ActorName:
                    Found=1
                    break
            if not Found:
                self.PartBox.insert(END,ActorName)

        # Force all parts to be highlighted, so you can unhighlight
        # only the BODY or whatever.  (Better than my usual leave-it-alone
        # to select all.)
        self.PartBox.selection_set(0,END)

        self.ShowStatus("%d frames. Deselect unwanted parts, set values, hit Go." % FileFrameCount)
        return 1

    # - - - - - - - - - - - - - - - - - -

    def Action(self):
        # Take the array of names and values from the BVH,
        # and apply them in "modulated" form.  Specifically,
        # take the diff btn the BVH value and current value,
        # and set the new value to an interpolated point between
        # its current and the BVH.
        self.ShowStatus("Turning off IK...")
        for OneIK in range(TheFig.NumIkChains()):
            TheFig.SetIkStatus(OneIK,0)

        # On the tran and rot: If you leave blank, we force
        # the values to the BVH exactly.  This is NOT the
        # same as putting a 1.0 in the blanks.

        RotString= self.RotScaleEntry.get()
        if RotString:
            RotScale=float(RotString)
            UseRot=1
        else:
            RotScale=1.0
            UseRot=0

        TrnString= self.TrnScaleEntry.get()
        if TrnString:
            TrnScale=float(TrnString)
            UseTrn=1
        else:
            TrnScale=1.0
            UseTrn=0

        SFString= self.StartEntry.get()
        if SFString:
            StartFrame=int(SFString)
        else:
            StartFrame=1

        # Repeats:
        RepString=self.RepEntry.get()
        if RepString:
            Repeats=int(RepString)
        else:
            Repeats=1

        EndFrame = StartFrame + Repeats*FileFrameCount

        # Increase the frames by using a fake PZ2 file.
        if EndFrame > scene.NumFrames():
            self.ShowStatus("Increasing frame count...")
            if not FakeFrames(EndFrame):
                self.ShowStatus("Failed to increase frames.")
                return

        Frame=StartFrame
        scene.SetFrame(Frame-1)
        scene.DrawAll()

        # Run through one iteration.  If Repeats was > 1, this will happen more than once.
        for dummyindex in range(Repeats):
            ValueIndex=0 # Start at top of Values for each repeat
            while Frame < EndFrame:
                scene.SetFrame(Frame-1)
                self.ShowStatus("Frame %d" % Frame)
                for (ActorName,ParmCode,ParmType) in Guide:
                    OneValue=Values[ValueIndex] # Take Values straight through, during one repeat.
                    ValueIndex=ValueIndex+1
                    # Note that the ValueIndex has to move for each entry in Guide, even
                    # when the actor isn't found.
                    try:
                        A=TheFig.Actor(ActorName)
                    except: # Just skip parts that aren't in this figure
                        continue
                    Parm=A.ParameterByCode(ParmCode)
                    try:
                        Parm.AddKeyFrame(Frame-1)
                    except:
                        continue


                    if ParmType==1: # Rot
                        if UseRot:
                            Delta=(OneValue-Parm.Value()) * RotScale  # Interpolate
                            Parm.SetValue(Parm.Value() + Delta)
                        else:
                            Parm.SetValue(OneValue)
                    else: # Tran or scale
                        OneValue = OneValue * MainScaler # Apply the standard multiplier
                        if UseTrn:
                            Delta=(OneValue-Parm.Value()) * TrnScale  # Interpolate
                            Parm.SetValue(Parm.Value() + Delta)
                        else:
                            Parm.SetValue(OneValue)
                if self.FloorVar.get(): TheFig.DropToFloor()
                Frame=Frame+1
        self.ShowStatus("Done.")

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def HandleGo(self):
        self.Action()
        #self.master.destroy()

    # - - - - - - - - - - - - - - - - - -

    def HandleCancel(self):
        self.master.destroy()

    # - - - - - - - - - - - - - - - - - -

    def Update(self):
        if scene: scene.ProcessSomeEvents(1)
        root.lift()
        root.after(250, self.Update)

#------------------------------------------------------------
# Activate the loop
#------------------------------------------------------------

app = App(root, '')
app.Update()
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
